import aiohttp.web

from .main import AsyncMiniGram


class AioMiniGram(AsyncMiniGram):
    async def aiohttp_handler(
        self, request: aiohttp.web.Request
    ) -> aiohttp.web.Response:
        try:
            data = await request.json()
            await self.async_handler(data)
        except Exception as e:
            print("Error in http_handler, aiohttp", e)
        return aiohttp.web.Response()



import asyncio
import json
import re
import time
from copy import deepcopy
from typing import Any

from .request import async_req, sync_req

DEBUG = False
ALLOWED_UPDATES = [
    "message",
    "edited_message",
    "channel_post",
    "edited_channel_post",
    "inline_query",
    "chosen_inline_result",
    "callback_query",
    "shipping_query",
    "pre_checkout_query",
    "poll",
    "poll_answer",
    "my_chat_member",
    "chat_member",
    "chat_join_request",
    "message_reaction",
    "message_reaction_count",
    "chat_boost",
    "removed_chat_boost",
]

# ANSI color codes
RED = "\033[91m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
BLUE = "\033[94m"
RESET = "\033[0m"  # Reset to default color


def parse_markdown(text: str, p: bool = False) -> str:
    text = re.sub(r"&", "&amp;", text)
    text = re.sub(r"<", "&lt;", text)
    text = re.sub(r">", "&gt;", text)

    patterns = [
        (r'(http[s]?://[^(\s|")]+)', '<a href="\\1">\\1</a>'),
        (r"^# (.*)$", "<h1>\\1</h1>"),
        (r"\*\*(.+?)\*\*", "<b>\\1</b>"),
        (r"__(.+?)__", "<i>\\1</i>"),
        (r"```(\w+)\n(.*?)```", '<pre><code class="language-\\1">\\2</code></pre>'),
        (r"```\n(.*?)```", "<pre><code>\\1</code></pre>"),
        (r"`([^`]+)`", "<code>\\1</code>"),
        (r"^> (.*)", "<blockquote>\\1</blockquote>"),
    ]
    for pattern, replacement in patterns:
        text = re.sub(pattern, replacement, text, flags=re.MULTILINE | re.DOTALL)

    if p:
        lines = text.split("\n")
        in_code = False
        for i, line in enumerate(lines):
            line = line.strip()
            if line.startswith("<pre"):
                in_code = True
            if "</pre>" in line:
                in_code = False
                continue
            if not in_code:
                lines[i] = "<p>" + line + "</p>"

        text = "\n".join(lines)

    return text


def extract_value(data: dict[str, Any], dotted: str, default: Any = None) -> Any:
    if "." not in dotted:
        return data.get(dotted, default)
    segment, other = dotted.split(".", 1)
    if segment not in data:
        return default
    return extract_value(data[segment], other)


def now():
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def debug(color, title, data):
    if DEBUG:
        print(f"\n{color}{now()} - {title}{RESET}\n{json.dumps(data, indent=2)}")


def dprint(*args, **kwargs):
    if DEBUG:
        print(*args, **kwargs)


class MiniGramUpdate:
    def __init__(self, data: dict, need_payload=False):
        pl = {}
        self.update_type = "unknown"
        for update_type in ALLOWED_UPDATES:
            if update_type in data:
                pl = deepcopy(data[update_type])
                self.update_type = update_type
                break
        if not pl:
            raise ValueError("unknown or unallowed update_type")

        self.update_id = extract_value(data, "update_id")
        self.chat_id = extract_value(pl, "chat.id")
        self.text = extract_value(pl, "text")
        self.message_id = extract_value(pl, "message_id")
        self.from_user = extract_value(pl, "from", {})
        self.user = extract_value(pl, "user", {})
        if len(self.from_user) > 0:
            self.from_id = self.from_user.get("id")
        else:
            self.from_id = self.user.get("id")

        debug(RED, "raw update:", data)

        self.payload: dict | None = pl if need_payload else None
        dprint(self)

    def __repr__(self):
        return f"<MiniGramUpdate with type `{self.update_type}` from {self.from_id}/{self.from_user} {self.chat_id} {self.message_id} {self.text}>"


class BaseMiniGram:
    def __init__(self, key: str, need_payload=False):
        self.key = key
        self.need_payload = need_payload
        self.last_updated_id = 0
        self.post_init()
        self.allowed_updates = ALLOWED_UPDATES
        dprint(f"Start time: {now()}\nDEBUG is {DEBUG}")

    def post_init(self):
        pass


class AsyncMiniGram(BaseMiniGram):
    def __init__(self, key: str, need_payload=False):
        global CURRENT_ASYNC
        super().__init__(key, need_payload)
        CURRENT_ASYNC = self

    async def shutdown(self):
        await self.delete_webhook()

    @classmethod
    def current(cls) -> "AsyncMiniGram":
        global CURRENT_ASYNC
        if CURRENT_ASYNC is None:
            raise ValueError("No current MiniGram")
        return CURRENT_ASYNC

    async def req(self, method: str, **kwargs) -> dict:
        if method != "getUpdates" and DEBUG:
            tmp = deepcopy(kwargs)
            tmp["method"] = method
            debug(YELLOW, "send request:", tmp)

        url = f"https://api.telegram.org/bot{self.key}/{method}"
        code, response = await async_req(url, kwargs)

        if method != "getUpdates" and DEBUG:
            tmp = deepcopy(response)
            tmp["code"] = code
            debug(GREEN, "get response:", tmp)

        return response

    async def get_updates(self) -> dict:
        params = {"timeout": 60, "allowed_updates": self.allowed_updates}
        if self.last_updated_id != 0:
            params["offset"] = self.last_updated_id + 1
        return await self.req("getUpdates", **params)

    async def start_polling(self):
        while True:
            updates = await self.get_updates()
            for update in updates.get("result", []):
                if update["update_id"] > self.last_updated_id:
                    await self.handle_update(MiniGramUpdate(update, need_payload=self.need_payload))
                    self.last_updated_id = update["update_id"]
                else:
                    dprint(f"{RED}{now()} - skip update with update_id <= self.last_updated_id {RESET}")
            await asyncio.sleep(0.1)

    async def handle_update(self, update: MiniGramUpdate):
        pass

    async def send_text(self, chat_id: int, text: str, parse_mode: str = "HTML", **kwargs) -> dict:
        return await self.req("sendMessage", chat_id=chat_id, text=text, parse_mode=parse_mode, **kwargs)

    async def reply(self, update: MiniGramUpdate, text: str = "", **kwargs) -> dict:
        params = {"chat_id": update.chat_id, "reply_to_message_id": update.message_id, "text": text, **kwargs}
        return await self.req("sendMessage", **params)

    async def send_chat_action(self, chat_id, action, **kwargs):
        if isinstance(chat_id, MiniGramUpdate):
            chat_id = chat_id.chat_id
        return await self.req("sendChatAction", chat_id=chat_id, action=action, **kwargs)

    async def set_message_reaction(self, update: MiniGramUpdate, reaction: str = "", is_big=False):
        params = {
            "chat_id": update.chat_id,
            "message_id": update.message_id,
            "reaction": [{"type": "emoji", "emoji": reaction}],
            "is_big": is_big,
        }
        return await self.req("setMessageReaction", **params)

    async def set_webhook(self, url: str) -> dict:
        return await self.req("setWebhook", url=url)

    async def delete_webhook(self) -> dict:
        return await self.req("deleteWebhook")

    async def get_webhook_info(self) -> dict:
        return await self.req("getWebhookInfo")

    async def async_handler(self, data: dict) -> None:
        await self.handle_update(MiniGramUpdate(data))


class MiniGram(BaseMiniGram):
    def __init__(self, key: str, need_payload=False):
        global CURRENT_SYNC
        super().__init__(key, need_payload=need_payload)
        CURRENT_SYNC = self

    def shutdown(self):
        self.delete_webhook()

    @classmethod
    def current(cls) -> "MiniGram":
        global CURRENT_SYNC
        if CURRENT_SYNC is None:
            raise ValueError("No current MiniGram")
        return CURRENT_SYNC

    def req(self, method: str, **kwargs) -> dict:
        if method != "getUpdates" and DEBUG:
            tmp = deepcopy(kwargs)
            tmp["method"] = method
            debug(YELLOW, "send request:", tmp)

        url = f"https://api.telegram.org/bot{self.key}/{method}"

        code, response = sync_req(url, kwargs)

        if method != "getUpdates" and DEBUG:
            tmp = deepcopy(response)
            tmp["code"] = code
            debug(GREEN, "get response:", tmp)

        return response

    def get_updates(self) -> dict:
        params = {"timeout": 60, "allowed_updates": self.allowed_updates}
        if self.last_updated_id != 0:
            params["offset"] = self.last_updated_id + 1
        return self.req("getUpdates", **params)

    def start_polling(self):
        while True:
            updates = self.get_updates()
            for update in updates.get("result", []):
                if update["update_id"] > self.last_updated_id:
                    self.handle_update(MiniGramUpdate(update, need_payload=self.need_payload))
                    self.last_updated_id = update["update_id"]
                else:
                    dprint(f"{RED}{now()} - skip update with update_id <= self.last_updated_id {RESET}")
            time.sleep(0.01)

    def handle_update(self, update: MiniGramUpdate):
        pass

    def send_text(self, chat_id: int, text: str, parse_mode: str = "HTML", **kwargs) -> dict:
        return self.req("sendMessage", chat_id=chat_id, text=text, parse_mode=parse_mode, **kwargs)

    def reply(self, update: MiniGramUpdate, text: str = "", **kwargs) -> dict:
        params = {"chat_id": update.chat_id, "reply_to_message_id": update.message_id, "text": text, **kwargs}
        return self.req("sendMessage", **params)

    def send_chat_action(self, chat_id: Any, action, **kwargs):
        if isinstance(chat_id, MiniGramUpdate):
            chat_id = chat_id.chat_id
        return self.req("sendChatAction", chat_id=chat_id, action=action, **kwargs)

    def set_message_reaction(self, update: MiniGramUpdate, reaction: str = "", is_big=False):
        params = {
            "chat_id": update.chat_id,
            "message_id": update.message_id,
            "reaction": [{"type": "emoji", "emoji": reaction}],
            "is_big": is_big,
        }
        return self.req("setMessageReaction", **params)

    def set_webhook(self, url: str) -> dict:
        return self.req("setWebhook", url=url)

    def delete_webhook(self) -> dict:
        return self.req("deleteWebhook")

    def get_webhook_info(self) -> dict:
        return self.req("getWebhookInfo")

    def sync_handler(self, data: dict) -> None:
        self.handle_update(MiniGramUpdate(data))


CURRENT_ASYNC: AsyncMiniGram | None = None
CURRENT_SYNC: MiniGram | None = None



from fastapi import Request
from fastapi.responses import JSONResponse
from .main import AsyncMiniGram

class FastAPIMiniGram(AsyncMiniGram):
    async def fastapi_handler(self, request: Request) -> JSONResponse:
        try:
            data = await request.json()
            await self.async_handler(data)
        except Exception as e:
            print("Error in http_handler, fastapi", e)
        return JSONResponse({"status": "ok"})



import importlib.util

from .main import MiniGram, MiniGramUpdate, AsyncMiniGram

__all__ = ["MiniGram", "MiniGramUpdate", "AsyncMiniGram"]

if importlib.util.find_spec("aiohttp"):
    from .aiohttp import AioMiniGram  # noqa

    __all__.append("AioMiniGram")

if importlib.util.find_spec("starlette"):
    from .starlette import StarletteMiniGram  # noqa

    __all__.append("StarletteMiniGram")



from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from .main import AsyncMiniGram


class StarletteMiniGram(AsyncMiniGram):
    async def starlette_handler(self, request: Request) -> Response:
        try:
            data = await request.json()
            await self.async_handler(data)
        except Exception as e:
            print("Error in http_handler, starlette", e)
        return JSONResponse({"status": "ok"})



import asyncio
import asyncio.exceptions
import importlib.util
import json
import ssl
import urllib.error
import urllib.request
from typing import Dict, Optional, Tuple
from urllib.parse import urlparse

if importlib.util.find_spec("aiohttp"):
    import aiohttp

    async def async_req(url: str, data: Optional[dict] = None) -> Tuple[int, dict]:
        try:
            async with aiohttp.ClientSession() as client:
                async with client.post(url, json=data, headers={"content-type": "application/json"}) as response:
                    return response.status, await response.json()
        except asyncio.exceptions.TimeoutError:
            return 200, {"result": []}

elif importlib.util.find_spec("httpx"):
    import httpx

    async def async_req(url: str, data: Optional[dict] = None) -> Tuple[int, dict]:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    url,
                    json=data,
                    headers={"content-type": "application/json"},
                    timeout=180,
                )
                return response.status_code, response.json()
        except httpx.ReadTimeout:
            return 200, {"result": []}

    def sync_req(url: str, data: Optional[dict] = None) -> Tuple[int, dict]:
        try:
            with httpx.Client() as client:
                response = client.post(
                    url,
                    json=data,
                    headers={"content-type": "application/json"},
                    timeout=180,
                )
                return response.status_code, response.json()
        except httpx.ReadTimeout:
            return 200, {"result": []}

else:

    async def async_req(url: str, data: Optional[Dict] = None) -> Tuple[int, Dict]:
        parsed_url = urlparse(url)
        host, port = (
            parsed_url.hostname,
            parsed_url.port or (443 if parsed_url.scheme == "https" else 80),
        )

        headers = {"Host": host, "Content-Type": "application/json"}

        if data is None:
            data = {}

        body = json.dumps(data).encode("utf-8")
        content_length = len(body)
        headers["Content-Length"] = str(content_length)

        rb = f"POST {parsed_url.path or '/'} HTTP/1.1\r\n"
        rb += "\r\n".join(f"{key}: {value}" for key, value in headers.items())
        rb += "\r\n\r\n"
        request = rb.encode("utf-8") + body

        scheme = ssl._create_unverified_context() if parsed_url.scheme == "https" else None
        reader, writer = await asyncio.open_connection(host, port, ssl=scheme)
        writer.write(request)
        await writer.drain()

        status_line = await reader.readline()
        status_code = int(status_line.split()[1])

        headers = {}
        while True:
            line = await reader.readline()
            if line == b"\r\n":
                break
            key, value = line.decode("utf-8").strip().split(": ", 1)
            headers[key] = value

        response_data = b""
        content_length = int(str(headers.get("Content-Length", "0")))
        while True:
            chunk = await reader.read(1024)
            response_data += chunk
            if len(response_data) >= content_length or not chunk:
                break

        writer.close()
        await writer.wait_closed()

        response_json = json.loads(response_data.decode("utf-8"))
        return status_code, response_json

    def sync_req(url: str, data: Optional[dict] = None) -> Tuple[int, dict]:
        headers = {"content-type": "application/json"}
        try:
            ssl_context = ssl._create_unverified_context()
            req = urllib.request.Request(url, data=json.dumps(data).encode("utf-8"), headers=headers)
            with urllib.request.urlopen(req, timeout=180, context=ssl_context) as response:
                status_code = response.getcode()
                response_data = response.read()
                return status_code, json.loads(response_data.decode("utf-8"))
        except urllib.error.URLError as e:
            return int(e.errno) if e.errno else 200, {"result": e.reason}
